# UI and connectSignals for MainWindow
from math import sqrt
import numpy as np

from PySide2 import QtWidgets, QtCore

from sohstationviewer.view.core import plotting_widget as plottingWidget
from sohstationviewer.controller.plottingData import (
    getTitle, getDayTicks, formatTime)
from sohstationviewer.model.handling_data import (
    get_trimTPSData, get_eachDay5MinList, findTPSTm)
from sohstationviewer.database.extractData import (
    getColorDef, getColorRanges, getChanLabel)
from sohstationviewer.conf.colorSettings import Clr
from sohstationviewer.conf import constants as const
from sohstationviewer.controller.util import displayTrackingInfo, fmti


class TimePowerSquaredWidget(plottingWidget.PlottingWidget):
    """
    Widget to display time power square data for waveform channels
    """
    def __init__(self, *args, **kwarg):
        """
        rulers: [matplotlib.lines.Line2D,] - list of squares on each waveform
            channels to highlight the five-minute at the mouse click on
            TimePowerSquaredWidget or the five-minute corresponding to the
            time at the mouse click on other plotting widgets
        """
        self.rulers = []
        """
        zoom_marker1s: [matplotlib.lines.Line2D,] - list of line markers to
            mark the five-minute corresponding to the start of the zoom area
        """
        self.zoom_marker1s = []
        """
        zoom_marker2s: [matplotlib.lines.Line2D,] - list of line markers to
            mark the five-minute corresponding to the end of the zoom area
        """
        self.zoom_marker2s = []
        """
        every_day_5_min_list: [[288 of floats], ] - the list of all starts
            of five minutes for every day in which each day has 288 of
            5 minutes.
        """
        self.each_day5_min_list = []

        super().__init__(*args, **kwarg)
        self.fig.canvas.mpl_connect('pick_event', self.on_pick_event)

    def plot_tps_channels(self, start_tm=None, end_tm=None, key=None,
                          data_time=None, waveform_data=None):
        """
        Recursively plot each channels for waveform_data and mass_pos_data.

        :param start_tm: float - requested start time to read
        :param end_tm: float - requested end time to read
        :param key: str or (str, str) station name  for mseed,
            or (device's experiment number, experiment number ) for reftek
        :param data_time: [float, float] - time range of the selected data set
        :param waveform_data: dict - read waveform data of selected data set,
            refer to DataTypeModel.__init__.waveform_data[key]['read_data']
        """
        self.processing_log = []  # [(message, type)]
        self.gap_bar = None
        self.min_x = max(data_time[0], start_tm)
        self.max_x = min(data_time[1], end_tm)

        if self.axes:
            self.fig.clear()
        self.date_mode = self.parent.date_format.upper()
        if waveform_data is not None:
            self.plotting_data1 = waveform_data
            self.min_x = max(data_time[0], start_tm)
            self.max_x = min(data_time[1], end_tm)
            self.plot_total = len(waveform_data)
        title = getTitle(key, self.min_x, self.max_x, self.date_mode)
        self.plotting_bot = const.BOTTOM
        self.plotting_bot_pixel = const.BOTTOM_PX
        self.axes = []
        self.rulers = []
        self.zoom_marker1s = []
        self.zoom_marker2s = []

        self.timestamp_bar_top = self.add_timestamp_bar(0.)
        self.set_title(title, y=0, v_align='bottom')
        self.each_day5_min_list = get_eachDay5MinList(self.min_x, self.max_x)
        for chanID in self.plotting_data1:
            ax = self.get_plot_data(self.plotting_data1[chanID], chanID)
            self.axes.append(ax)
        self.set_legend()
        # Set view size fit with the given data
        if self.main_widget.geometry().height() < self.plotting_bot_pixel:
            self.main_widget.setFixedHeight(self.plotting_bot_pixel)
        self.set_lim_markers()
        self.draw()

    def get_plot_data(self, c_data, chan_id):
        """
        Trim data to minx, max_x and calculate time-power-square for each 5
            minute into c_data['tps_data'] then draw each 5 minute with the
            color corresponding to value
        Create ruler, zoom_marker1, zoom_marker2 for the channel.

        :param c_data: dict - data of the channel which includes down-sampled
            data in keys 'times' and 'data'. Refer to
            DataTypeModel.__init__.waveform_data[key]['read_data'][chan_id]
        :param chan_id: str - name of channel
        :return ax: matplotlib.axes.Axes - axes of the channel
        """
        if 'tps_data' not in c_data:
            # get new minX, maxX according to exact start time  of days
            get_trimTPSData(c_data, self.min_x, self.max_x,
                            self.each_day5_min_list)
        total_days = c_data['tps_data'].shape[0]
        plot_h = self.get_height(1.5 * total_days, bw_plots_distance=0.003)
        ax = self.create_axes(self.plotting_bot, plot_h)
        ax.text(
            -0.1, 1.2,
            f"{getChanLabel(chan_id)} {c_data['samplerate']}",
            horizontalalignment='left',
            verticalalignment='top',
            rotation='horizontal',
            transform=ax.transAxes,
            color=self.display_color['plot_label'],
            size=self.font_size + 2
        )

        zoom_marker1 = ax.plot(
            [], [], marker='|', markersize=10,
            markeredgecolor=self.display_color['zoom_marker'])[0]
        self.zoom_marker1s.append(zoom_marker1)

        zoom_marker2 = ax.plot(
            [], [], marker='|', markersize=10,
            markeredgecolor=self.display_color['zoom_marker'])[0]
        self.zoom_marker2s.append(zoom_marker2)

        ruler = ax.plot(
            [], [], marker='s', markersize=5,
            markeredgecolor=self.display_color['time_ruler'],
            markerfacecolor='None')[0]
        self.rulers.append(ruler)

        x = np.array([i for i in range(const.NO_5M_DAY)])
        square_counts = self.parent.sel_square_counts  # square counts range
        color_codes = self.parent.color_def  # colordef

        for dayIdx, y in enumerate(c_data['tps_data']):
            # not draw data out of day range
            color_set = self.get_color_set(y, square_counts, color_codes)
            # (- dayIdx): each day is a line, increase from top to bottom
            ax.scatter(x, [- dayIdx] * len(x), marker='|',
                       c=color_set, s=7, alpha=0.8)
        # extra to show highlight square
        ax.set_ylim(-(c_data['tps_data'].shape[0] + 1), 1)

        return ax

    def set_legend(self):
        """
        Plot one dot for each color and assign label to it. The dots are
            plotted outside of xlim to not show up in plotting area. xlim is
            set so that it has some extra space to show full hightlight square
            of the ruler.
        ax.legend will create one label for each dot.
        """
        # set height of legend and distance bw legend and upper ax
        plot_h = self.get_height(7, bw_plots_distance=0.003)
        ax = self.canvas.figure.add_axes(
            [self.plotting_l, self.plotting_bot, self.plotting_w, plot_h],
            picker=True
        )
        ax.patch.set_alpha(0)
        c_labels = self.parent.sel_col_labels
        clrs = self.parent.color_def  # colordef
        for idx in range(len(c_labels)):
            # draw a dot out of xlim so it isn't displayed in plotting area
            ax.scatter([300], [1],
                       c=Clr[clrs[idx]],
                       s=0.1,
                       label=c_labels[idx],
                       edgecolor=self.display_color['basic'],
                       alpha=0.8,
                       zorder=1,
                       picker=True)
        # extra to show highlight square
        ax.set_xlim(-2, const.NO_5M_DAY + 1)
        ax.legend(loc="upper left", framealpha=0.2,
                  markerscale=25,
                  labelcolor=self.display_color['basic'])

    def get_color_set(self, y, square_counts, colors):
        """
        Create array of color (col) according to value of y compare with
            square_counts (square count range)

        :param y: np.array of float - tps values for all 5-minutes of a day
        :param square_counts: [int, ] - list of square count ranges
        :param colors: [str,] - list of color codes based on colorSettings.Clr
        :return: np.array of color hex for each 5-minutes of a day
        """
        return (
            np.where(
                y == square_counts[0], Clr[colors[0]], np.where(
                    y < square_counts[1], Clr[colors[1]], np.where(
                        y < square_counts[2], Clr[colors[2]], np.where(
                            y < square_counts[3], Clr[colors[3]], np.where(
                                y < square_counts[4], Clr[colors[4]], np.where(
                                    y < square_counts[5], Clr[colors[5]], np.where(             # noqa: E501
                                        y < square_counts[6], Clr[colors[6]], Clr[colors[7]]    # noqa: E501
                                    )))))))
        )

    def create_axes(self, plot_b, plot_h, has_min_max_lines=False):
        """
        Create axes for 288 of 5m in a day in which minor tick for every hour,
            major tick for every 4 hour

        :param plot_b: float - bottom of the plot
        :param plot_h: float - height of the plot
        :param has_min_max_lines: bool - flag showing if the plot need min/max
            lines
        :return ax: matplotlib.axes.Axes - axes of tps of a waveform channel
        """
        ax = self.canvas.figure.add_axes(
            [self.plotting_l, plot_b, self.plotting_w, plot_h],
            picker=True
        )
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.xaxis.grid(True, which='major',
                      color=self.display_color['basic'], linestyle='-')
        ax.xaxis.grid(True, which='minor',
                      color=self.display_color['sub_basic'], linestyle='-')
        ax.set_yticks([])

        times, major_times, major_time_labels = getDayTicks()
        ax.set_xticks(times, minor=True)
        ax.set_xticks(major_times)
        ax.set_xticklabels(major_time_labels, fontsize=self.font_size,
                           color=self.display_color['basic'])
        # extra to show highlight square
        ax.set_xlim(-2, const.NO_5M_DAY + 1)
        ax.patch.set_alpha(0)
        return ax

    def on_pick_event(self, event):
        """
        When a plot is select, the corresponding point on each plot will
            be highlighted and their time and counts will be displayed.
        :param event: pick event - event when object of canvas is selected.
            The event happens before button_press_event.
        """
        info_str = ""
        if event.artist in self.axes:
            xdata = event.mouseevent.xdata
            if xdata is None:
                return
            xdata = round(xdata)
            # when click on outside xrange that close to edge, adjust to edge
            if xdata in [-2, -1]:
                xdata = 0
            if xdata in [288, 289]:
                xdata = 287
            ydata = round(event.mouseevent.ydata)

            y_idx = - ydata
            x_idx = xdata
            # identify time for rulers on other plotting widget
            self.tps_t = self.each_day5_min_list[y_idx, x_idx]
            format_t = formatTime(self.tps_t, self.date_mode, 'HH:MM:SS')
            info_str += f"{format_t}:"
            for chanID in self.plotting_data1:
                c_data = self.plotting_data1[chanID]
                data = c_data['tps_data'][y_idx, x_idx]
                info_str += f"  {chanID}:{fmti(sqrt(data))}"
            info_str += "  (counts)"
            displayTrackingInfo(self.tracking_box, info_str)
            self.draw()

    def on_ctrl_cmd_click(self, xdata):
        """
        Ctrl + cmd: base on xdata to find indexes of x an y
            in self.each_day_5_min_list to display ruler for each channel, then
            set time for each ruler in self.rulers to place them in correct
            position.

        :param xdata: float - time value in other plot
        """
        self.zoom_marker1_shown = False
        x_idx, y_idx = findTPSTm(xdata, self.each_day5_min_list)
        for rl in self.rulers:
            rl.set_data(x_idx, y_idx)

    def on_shift_click(self, xdata):
        """
        Shift + right click on other plot widget, this function will be called
            to show marker for place when it is zoomed.
            On the fist of zoom_marker, make ruler disappeared, set min_x.
            On the second of zoom_maker, call set_lim_markers to mark the new
                limit.

        :param xdata: float - time value in other plot
        """

        if not self.zoom_marker1_shown:
            self.set_rulers_invisible()
            self.min_x = xdata
            self.zoom_marker1_shown = True
        else:
            [self.min_x, self.max_x] = sorted(
                [self.min_x, xdata])
            self.set_lim_markers()
            self.zoom_marker1_shown = False

    def set_rulers_invisible(self):
        """
        Clear data for self.rulers to make them disappeared.
        """
        for rl in self.rulers:
            rl.set_data([], [])

    def set_lim_markers(self):
        """
        Find x index (which index in five minutes of a day) and
            y index (which day) of self.min_x and self.min_y, and set data for
            all markers in self.zoom_marker1s and self.zoom_marker2s.
        """
        x_idx, y_idx = findTPSTm(self.min_x, self.each_day5_min_list)
        for zm1 in self.zoom_marker1s:
            zm1.set_data(x_idx, y_idx)
        x_idx, y_idx = findTPSTm(self.max_x, self.each_day5_min_list)
        for zm2 in self.zoom_marker2s:
            zm2.set_data(x_idx, y_idx)


class TimePowerSquaredDialog(QtWidgets.QWidget):
    def __init__(self, parent):
        """
        Dialog to display time power square data for waveform channels. Users
            are allowed to select different Color Ranges (antarctica, low,
            medium, high) for differences values vs. colors map to draw
            time power square for each 5-minute data

        :param parent: QMainWindow/QWidget - the parent widget
        """
        super().__init__()
        self.parent = parent
        """
        date_format: str - format to display date/time
        """
        self.date_format = self.parent.date_format
        """
        data_type: str - type of data being plotted
        """
        self.data_type = None
        self.setGeometry(50, 50, 1200, 700)
        self.setWindowTitle("TPS Plot")

        main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(main_layout)
        main_layout.setContentsMargins(5, 5, 5, 5)
        main_layout.setSpacing(0)

        """
        trackingInfoTextBrowser: QTextBrowser - to display info text
        """
        self.info_text_browser = QtWidgets.QTextBrowser(self)
        """
        plotting_widget: PlottingWidget - the widget to draw time-power-square
            for each 5-minute of data
        """
        self.plotting_widget = TimePowerSquaredWidget(
            self, self.info_text_browser, "timepowersquaredwidget")
        main_layout.addWidget(self.plotting_widget, 2)

        bottom_layout = QtWidgets.QHBoxLayout()
        bottom_layout.addSpacing(20)
        main_layout.addLayout(bottom_layout)

        # ################ Color range #################
        bottom_layout.addWidget(QtWidgets.QLabel("Color Range"))

        """
        color_def: [str,] -  list of color codes in order of values to be
            displayed
        """
        self.color_def = getColorDef()
        """
        sel_square_counts: [int,] - selected time-power-square ranges
        """
        self.sel_square_counts = []
        """
        sel_col_labels: [str,] - color label for each range
        """
        self.sel_col_labels = []
        """
        color_ranges: [str,] - name of color:value map for user to choose
            'antarctica'/'low'/'med'/'high'
        all_square_counts: [[int,],] - time-power-squared values
        color_label: [str,] - labels that define count range for colors to show
            in the legend of the tps plotting widget
        """
        (self.color_ranges,
         self.all_square_counts,
         self.color_label) = getColorRanges()

        """
        color_range_choice: QComboBox - dropdown box for user to choose a
            color:value map name
        """
        self.color_range_choice = QtWidgets.QComboBox(self)
        self.color_range_choice.addItems(self.color_ranges)
        self.color_range_choice.setCurrentText('high')
        bottom_layout.addWidget(self.color_range_choice)
        ################################################
        self.save_button = QtWidgets.QPushButton('Save', self)
        bottom_layout.addWidget(self.save_button)

        self.info_text_browser.setFixedHeight(60)
        bottom_layout.addWidget(self.info_text_browser)

        self.connect_signals()
        self.color_range_changed()

    def set_data(self, data_type, file_name):
        """
        Set data_type and the window's title.

        :param data_type: str - data type of data being plotted
        :param file_name: str - name of the file/folder of the data set to be
            displayed
        """
        self.data_type = data_type
        self.setWindowTitle("TPS Plot %s - %s" % (data_type, file_name))

    def resizeEvent(self, event):
        """
        When TimePowerDialog is resized, its plotting_widget need to initialize
            its size to fit the viewport.

        :param event: QResizeEvent - resize event
        """
        self.plotting_widget.init_size()

    def connect_signals(self):
        """
        Connect functions to widgets
        """
        self.save_button.clicked.connect(self.save)
        self.color_range_choice.currentTextChanged.connect(
            self.color_range_changed)

    @QtCore.Slot()
    def color_range_changed(self):
        """
        When selecting a different color range from the drop down menu, set
            new value for sel_square_counts and sel_col_labels
        """
        color_range = self.color_range_choice.currentText()
        cr_index = self.color_ranges.index(color_range)
        self.sel_square_counts = self.all_square_counts[cr_index]
        self.sel_col_labels = self.color_label[cr_index]

    @QtCore.Slot()
    def save(self):
        """
        Save the plotting to a file
        """
        print("save")
